from typing import Union, TypeVar, Callable, Tuple, Iterable

import tensorflow as tf


T = TypeVar('T')
Specification = Union[tf.SparseTensorSpec, tf.RaggedTensorSpec, tf.TensorSpec]
Structure = Union[T, Iterable['Structure']]  # type: ignore
TupleStructure = Union[T, Tuple['TupleStructure', ...]]  # type: ignore


class Schema:
    def __init__(self, specification: Structure[Specification]) -> None:
        self.specification = specification

    @classmethod
    def transform(
            cls,
            specification: Structure[Specification],
            transformer: Callable[[tf.TensorSpec], T]) -> TupleStructure[T]:

        if isinstance(specification, Iterable):
            return tuple(cls.transform(item, transformer) for item in specification)
        else:
            return transformer(specification)

    def get_types(self) -> TupleStructure[tf.DType]:
        return self.transform(self.specification, lambda specification: specification.dtype)

    def get_shapes(self) -> TupleStructure[tf.TensorShape]:
        return self.transform(self.specification, lambda specification: specification.shape)



import math

import tensorflow as tf


def tile_dimension(tensor: tf.Tensor, multiple: int, axis: int = -1) -> tf.Tensor:
    rank = tf.rank(tensor)

    axis = axis if axis >= 0 else rank + axis + 1

    multiples = tf.ones(rank + 1, dtype=tf.int32)
    multiples = tf.tensor_scatter_nd_update(multiples, [[axis]], [multiple])

    tensor_tiled = tf.tile(tf.expand_dims(tensor, axis), multiples)

    return tensor_tiled


def flatten(tensor: tf.Tensor, axis: int = -1) -> tf.Tensor:
    start_axis = axis - 1

    shape = tf.shape(tensor)
    shape = tf.concat((shape[:start_axis], [tf.reduce_prod(shape[start_axis:])]), axis=0)
    tensor_flattened = tf.reshape(tensor, shape)

    joined_shape = tensor.shape[start_axis:]

    if None not in joined_shape:
        shape = tensor.shape[:start_axis] + [math.prod(joined_shape)]
        tensor_flattened.set_shape(shape)

    return tensor_flattened



import typing
from abc import ABC, abstractmethod
from typing import Tuple, Optional, Generator, TypeVar, Union

import numpy as np


T = TypeVar('T')
TensorTuple = Union[np.ndarray, Tuple[np.ndarray, ...]]
BatchType = Union[
    Tuple[TensorTuple, None],
    Tuple[TensorTuple, TensorTuple],
    Tuple[TensorTuple, TensorTuple, np.ndarray],
]


class Dataset(ABC):
    size: int
    targets: Optional[np.ndarray] = None

    @abstractmethod
    @typing.no_type_check
    def get_batches(self, *args) -> Generator[BatchType, None, None]:
        ...

    @abstractmethod
    def subset(self: T, indices: np.ndarray) -> T:
        ...



from enum import Enum
from typing import Mapping, Union, MutableMapping, Collection, Tuple, List, Optional, NamedTuple

import numpy as np

from sfu_ml_lib.base.metrics import NumericType, AggregatorType, MetricKeyType


class Statistic(Enum):
    TARGET = 'Target'
    PREDICTION = 'Prediction'
    PROBABILITY = 'Probability'
    COST = 'Cost'
    BETA = 'Beta'
    LOGITS = 'Logits'
    WEIGHTS = 'Weights'


StatisticKey = Union[str, Statistic]
StatisticValue = Union[List[NumericType], NumericType, np.ndarray]


class Aggregation(NamedTuple):
    metric: MetricKeyType
    statistics: Collection[StatisticKey]
    function: AggregatorType


class MetricAggregator:
    def __init__(self, aggregators: Optional[Collection[Aggregation]] = None):
        self.aggregators: MutableMapping[MetricKeyType, Tuple[Collection[StatisticKey], AggregatorType]] = {}
        self.state: MutableMapping[StatisticKey, List[NumericType]] = {}

        if aggregators:
            self.register_aggregators(aggregators)

    @staticmethod
    def to_numeric_list(value: StatisticValue) -> List[NumericType]:
        if isinstance(value, np.ndarray):
            return value.tolist()
        elif isinstance(value, List):
            return value
        else:
            return [value]

    @staticmethod
    def to_str(value: Union[Enum, str]) -> str:
        return str(value.value) if isinstance(value, Enum) else value

    def register_aggregator(
            self,
            metric: MetricKeyType,
            statistics: Collection[StatisticKey],
            function: AggregatorType,
    ) -> None:

        self.aggregators[metric] = (statistics, function)

        for statistic in statistics:
            if statistic not in self.state:
                self.state[statistic] = []

    def register_aggregators(self, aggregators: Collection[Aggregation]):
        for aggregator in aggregators:
            self.register_aggregator(aggregator.metric, aggregator.statistics, aggregator.function)

    def add_statistic(self, key: StatisticKey, value: StatisticValue) -> None:
        self.state[key].extend(self.to_numeric_list(value))

    def add_statistics(self, statistics: Mapping[StatisticKey, StatisticValue]) -> None:
        for key, value in statistics.items():
            self.add_statistic(key, value)

    def compute_metric(self, key: MetricKeyType) -> Optional[NumericType]:
        statistics, function = self.aggregators[key]

        if not any(len(self.state[statistic]) != 0 for statistic in statistics):
            return None

        return function(*(np.array(self.state[statistic]) for statistic in statistics))

    def compute_metrics(self, keys: Collection[MetricKeyType]) -> Mapping[MetricKeyType, NumericType]:
        return {
            key: value
            for key, value
            in ((key, self.compute_metric(key)) for key in keys)
            if value is not None
        }

    def get_metrics(self) -> Mapping[MetricKeyType, NumericType]:
        return self.compute_metrics(self.aggregators.keys())

    def clear(self) -> None:
        for sequence in self.state.values():
            sequence.clear()

    def flush(self) -> Mapping[MetricKeyType, NumericType]:
        metrics = self.compute_metrics(self.aggregators.keys())

        self.clear()

        return metrics



from typing import List, Optional, Mapping, Union

import tensorflow as tf
from tensorflow import keras

from sfu_ml_lib.base.metrics import Metric
from sfu_ml_lib.metrics.error import CutoffThreshold


class F1Score(keras.metrics.Metric):
    def __init__(self, num_classes: int, average: str = 'weighted') -> None:
        if average == 'macro':
            metric = Metric.MACRO_F1
        elif average == 'micro':
            metric = Metric.MICRO_F1
        else:
            metric = Metric.WEIGHTED_F1

        super().__init__(name=metric.value, dtype=tf.float32)

        self.num_classes = num_classes
        self.average = average

        self.axis = 0 if self.average != 'micro' else None
        self.init_shape = [self.num_classes] if self.average != 'micro' else []

        self.true_positives = self.create_zero_weights('true_positives', self.init_shape)
        self.false_positives = self.create_zero_weights('false_positives', self.init_shape)
        self.false_negatives = self.create_zero_weights('false_negatives', self.init_shape)
        self.weights_intermediate = self.create_zero_weights('weights_intermediate', self.init_shape)

    def create_zero_weights(self, name: str, shape: List[tf.DType]) -> tf.Variable:
        return self.add_weight(name, shape=shape, initializer=tf.zeros_initializer, dtype=tf.float32)

    def compute_weighted_sum(self, values: tf.Tensor, sample_weight: tf.Tensor) -> tf.Tensor:
        if sample_weight is not None:
            values *= sample_weight[:, tf.newaxis]

        weighted_sum = tf.math.reduce_sum(values, axis=self.axis)

        return weighted_sum

    def update_state(self, y_true: tf.Tensor, y_pred: tf.Tensor, sample_weight: Optional[tf.Tensor] = None) -> None:
        self.true_positives.assign_add(self.compute_weighted_sum(y_pred * y_true, sample_weight))
        self.false_positives.assign_add(self.compute_weighted_sum(y_pred * (1. - y_true), sample_weight))
        self.false_negatives.assign_add(self.compute_weighted_sum((1. - y_pred) * y_true, sample_weight))
        self.weights_intermediate.assign_add(self.compute_weighted_sum(y_true, sample_weight))

    def result(self) -> tf.Tensor:
        precision = tf.math.divide_no_nan(self.true_positives, self.true_positives + self.false_positives)
        recall = tf.math.divide_no_nan(self.true_positives, self.true_positives + self.false_negatives)

        mul_value = 2 * precision * recall
        add_value = precision + recall
        f1_score = tf.math.divide_no_nan(mul_value, add_value)

        if self.average == 'weighted':
            weights = tf.math.divide_no_nan(self.weights_intermediate, tf.reduce_sum(self.weights_intermediate))
            f1_score = tf.reduce_sum(weights * f1_score)
        else:
            f1_score = tf.reduce_mean(f1_score)

        return f1_score

    def get_config(self) -> Mapping[str, Union[tf.Tensor, int, bool, float, str]]:
        config = {'num_classes': self.num_classes, 'average': self.average}

        base_config = super().get_config()

        return {**base_config, **config}

    def reset_states(self) -> None:
        self.true_positives.assign(tf.zeros(self.init_shape, tf.float32))
        self.false_positives.assign(tf.zeros(self.init_shape, tf.float32))
        self.false_negatives.assign(tf.zeros(self.init_shape, tf.float32))
        self.weights_intermediate.assign(tf.zeros(self.init_shape, tf.float32))


class MultiClassF1Score(F1Score):
    def update_state(self, y_true: tf.Tensor, y_pred: tf.Tensor, sample_weight: Optional[tf.Tensor] = None) -> None:
        targets = tf.one_hot(y_true, self.num_classes)
        predictions = tf.one_hot(tf.argmax(y_pred, axis=-1), self.num_classes)

        super().update_state(targets, predictions, sample_weight)


class MultiLabelF1Score(F1Score):
    def __init__(
            self,
            num_classes: int,
            average: str = 'weighted',
            from_logits: bool = False,
            adapt_thresholds: bool = False) -> None:

        super().__init__(num_classes, average)

        self.from_logits = from_logits
        self.cutoff_threshold = CutoffThreshold() if adapt_thresholds else None

    def get_thresholds(
            self,
            y_true: tf.Tensor,
            y_pred: tf.Tensor,
            sample_weight: Optional[tf.Tensor] = None) -> tf.Tensor:

        if self.cutoff_threshold is None:
            thresholds = tf.fill((self.num_classes,), tf.convert_to_tensor(.5, dtype=tf.float32))
        else:
            self.cutoff_threshold.update_state(y_true, y_pred, sample_weight)
            thresholds = self.result()

        return thresholds

    def update_state(self, y_true: tf.Tensor, y_pred: tf.Tensor, sample_weight: Optional[tf.Tensor] = None) -> None:
        thresholds = self.get_thresholds(y_true, y_pred, sample_weight)

        predictions = tf.sigmoid(y_pred) if self.from_logits else y_pred
        predictions = tf.cast(predictions >= thresholds, dtype=tf.float32)

        super().update_state(y_true, predictions, sample_weight)



import tensorflow as tf


def sample_gumbel(shape: tf.Tensor, eps: float = 1e-20) -> tf.Tensor:
    return -tf.math.log(-tf.math.log(tf.random.uniform(shape) + eps) + eps)


def add_gumbel_sample(log_probabilities: tf.Tensor) -> tf.Tensor:
    return log_probabilities + sample_gumbel(tf.shape(log_probabilities))


def sample_categorical_log_probabilities(log_probabilities: tf.Tensor, axis: int = -1) -> tf.Tensor:
    return tf.argmax(add_gumbel_sample(log_probabilities), axis=axis)


def softmax_masked(tensor: tf.Tensor, mask: tf.Tensor) -> tf.Tensor:
    rank_difference = len(tensor.shape) - len(mask.shape)
    axis = -(rank_difference + 1)

    lengths = tf.reduce_sum(tf.cast(mask, dtype=tf.int32), axis=-1)
    lengths = tf.expand_dims(lengths, axis=-1)

    extra_dims = [1] * rank_difference
    mask_shape = tf.concat((tf.shape(mask), extra_dims), axis=0)
    lengths_shape = tf.concat((tf.shape(lengths), extra_dims), axis=0)

    mask = tf.reshape(mask, mask_shape)
    lengths = tf.reshape(lengths, lengths_shape)

    tensor = tf.where(mask, tensor, -1e9)
    tensor = tf.nn.softmax(tensor, axis=axis)
    tensor = tf.where(lengths > 0, tensor, 0.)

    return tensor



import sfu_ml_lib.metrics.aggregation.metric_functions as metric_functions
from sfu_ml_lib.base.metrics import Metric
from sfu_ml_lib.metrics.aggregation.aggregator import Statistic, Aggregation


MEAN_ABSOLUTE_ERROR = Aggregation(
    metric=Metric.MEAN_ABSOLUTE_ERROR,
    statistics=[Statistic.TARGET, Statistic.LOGITS],
    function=metric_functions.mean_absolute_error,
)

MEAN_SQUARE_ERROR = Aggregation(
    metric=Metric.MEAN_SQUARE_ERROR,
    statistics=[Statistic.TARGET, Statistic.LOGITS],
    function=metric_functions.mean_squared_error,
)

PEARSON_CORRELATION = Aggregation(
    metric=Metric.PEARSON_CORRELATION,
    statistics=[Statistic.TARGET, Statistic.LOGITS],
    function=metric_functions.pearson_correlation,
)

ACCURACY_MULTICLASS = Aggregation(
    metric=Metric.ACCURACY,
    statistics=[Statistic.TARGET, Statistic.LOGITS],
    function=lambda targets, logits: metric_functions.accuracy(
        targets=targets,
        predictions=metric_functions.logits_to_predictions_multiclass(logits),
    ),
)

ACCURACY_MULTILABEL = Aggregation(
    metric=Metric.ACCURACY,
    statistics=[Statistic.TARGET, Statistic.LOGITS],
    function=lambda targets, logits: metric_functions.accuracy(
        targets=targets,
        predictions=metric_functions.logits_to_predictions_multilabel(logits),
    ),
)

JACCARD_SCORE_MULTILABEL = Aggregation(
    metric=Metric.JACCARD_SCORE,
    statistics=[Statistic.TARGET, Statistic.LOGITS],
    function=lambda targets, logits: metric_functions.jaccard_score_multilabel(
        targets=targets,
        predictions=metric_functions.logits_to_predictions_multilabel(logits),
    ),
)

CROSS_ENTROPY_MULTICLASS = Aggregation(
    metric=Metric.CROSS_ENTROPY,
    statistics=[Statistic.TARGET, Statistic.LOGITS],
    function=metric_functions.cross_entropy_multiclass,
)

CROSS_ENTROPY_MULTILABEL = Aggregation(
    metric=Metric.CROSS_ENTROPY,
    statistics=[Statistic.TARGET, Statistic.LOGITS],
    function=metric_functions.cross_entropy_multilabel,
)

WEIGHTED_F1_MULTICLASS = Aggregation(
    metric=Metric.WEIGHTED_F1,
    statistics=[Statistic.TARGET, Statistic.LOGITS],
    function=lambda targets, logits: metric_functions.f1_score(
        targets=targets,
        predictions=metric_functions.logits_to_predictions_multiclass(logits),
        average='weighted',
    ),
)

WEIGHTED_F1_MULTILABEL = Aggregation(
    metric=Metric.WEIGHTED_F1,
    statistics=[Statistic.TARGET, Statistic.LOGITS],
    function=lambda targets, logits: metric_functions.f1_score(
        targets=targets,
        predictions=metric_functions.logits_to_predictions_multilabel(logits),
        average='weighted',
    ),
)

WEIGHTED_PRECISION_MULTICLASS = Aggregation(
    metric=Metric.WEIGHTED_PRECISION,
    statistics=[Statistic.TARGET, Statistic.LOGITS],
    function=lambda targets, logits: metric_functions.precision(
        targets=targets,
        predictions=metric_functions.logits_to_predictions_multiclass(logits),
        average='weighted',
    ),
)

WEIGHTED_PRECISION_MULTILABEL = Aggregation(
    metric=Metric.WEIGHTED_PRECISION,
    statistics=[Statistic.TARGET, Statistic.LOGITS],
    function=lambda targets, logits: metric_functions.precision(
        targets=targets,
        predictions=metric_functions.logits_to_predictions_multilabel(logits),
        average='weighted',
    ),
)

WEIGHTED_RECALL_MULTICLASS = Aggregation(
    metric=Metric.WEIGHTED_RECALL,
    statistics=[Statistic.TARGET, Statistic.LOGITS],
    function=lambda targets, logits: metric_functions.recall(
        targets=targets,
        predictions=metric_functions.logits_to_predictions_multiclass(logits),
        average='weighted',
    ),
)

WEIGHTED_RECALL_MULTILABEL = Aggregation(
    metric=Metric.WEIGHTED_RECALL,
    statistics=[Statistic.TARGET, Statistic.LOGITS],
    function=lambda targets, logits: metric_functions.recall(
        targets=targets,
        predictions=metric_functions.logits_to_predictions_multilabel(logits),
        average='weighted',
    ),
)

MICRO_F1_MULTICLASS = Aggregation(
    metric=Metric.MICRO_F1,
    statistics=[Statistic.TARGET, Statistic.LOGITS],
    function=lambda targets, logits: metric_functions.f1_score(
        targets=targets,
        predictions=metric_functions.logits_to_predictions_multiclass(logits),
        average='micro',
    ),
)

MICRO_F1_MULTILABEL = Aggregation(
    metric=Metric.MICRO_F1,
    statistics=[Statistic.TARGET, Statistic.LOGITS],
    function=lambda targets, logits: metric_functions.f1_score(
        targets=targets,
        predictions=metric_functions.logits_to_predictions_multilabel(logits),
        average='micro',
    ),
)

MICRO_PRECISION_MULTICLASS = Aggregation(
    metric=Metric.MICRO_PRECISION,
    statistics=[Statistic.TARGET, Statistic.LOGITS],
    function=lambda targets, logits: metric_functions.precision(
        targets=targets,
        predictions=metric_functions.logits_to_predictions_multiclass(logits),
        average='micro',
    ),
)

MICRO_PRECISION_MULTILABEL = Aggregation(
    metric=Metric.MICRO_PRECISION,
    statistics=[Statistic.TARGET, Statistic.LOGITS],
    function=lambda targets, logits: metric_functions.precision(
        targets=targets,
        predictions=metric_functions.logits_to_predictions_multilabel(logits),
        average='micro',
    ),
)

MICRO_RECALL_MULTICLASS = Aggregation(
    metric=Metric.MICRO_RECALL,
    statistics=[Statistic.TARGET, Statistic.LOGITS],
    function=lambda targets, logits: metric_functions.recall(
        targets=targets,
        predictions=metric_functions.logits_to_predictions_multiclass(logits),
        average='micro',
    ),
)

MICRO_RECALL_MULTILABEL = Aggregation(
    metric=Metric.MICRO_RECALL,
    statistics=[Statistic.TARGET, Statistic.LOGITS],
    function=lambda targets, logits: metric_functions.recall(
        targets=targets,
        predictions=metric_functions.logits_to_predictions_multilabel(logits),
        average='micro',
    ),
)

MACRO_F1_MULTICLASS = Aggregation(
    metric=Metric.MACRO_F1,
    statistics=[Statistic.TARGET, Statistic.LOGITS],
    function=lambda targets, logits: metric_functions.f1_score(
        targets=targets,
        predictions=metric_functions.logits_to_predictions_multiclass(logits),
        average='macro',
    ),
)

MACRO_F1_MULTILABEL = Aggregation(
    metric=Metric.MACRO_F1,
    statistics=[Statistic.TARGET, Statistic.LOGITS],
    function=lambda targets, logits: metric_functions.f1_score(
        targets=targets,
        predictions=metric_functions.logits_to_predictions_multilabel(logits),
        average='macro',
    ),
)

MACRO_PRECISION_MULTICLASS = Aggregation(
    metric=Metric.MACRO_PRECISION,
    statistics=[Statistic.TARGET, Statistic.LOGITS],
    function=lambda targets, logits: metric_functions.precision(
        targets=targets,
        predictions=metric_functions.logits_to_predictions_multiclass(logits),
        average='macro',
    ),
)

MACRO_PRECISION_MULTILABEL = Aggregation(
    metric=Metric.MACRO_PRECISION,
    statistics=[Statistic.TARGET, Statistic.LOGITS],
    function=lambda targets, logits: metric_functions.precision(
        targets=targets,
        predictions=metric_functions.logits_to_predictions_multilabel(logits),
        average='macro',
    ),
)

MACRO_RECALL_MULTICLASS = Aggregation(
    metric=Metric.MACRO_RECALL,
    statistics=[Statistic.TARGET, Statistic.LOGITS],
    function=lambda targets, logits: metric_functions.recall(
        targets=targets,
        predictions=metric_functions.logits_to_predictions_multiclass(logits),
        average='macro',
    ),
)

MACRO_RECALL_MULTILABEL = Aggregation(
    metric=Metric.MACRO_RECALL,
    statistics=[Statistic.TARGET, Statistic.LOGITS],
    function=lambda targets, logits: metric_functions.recall(
        targets=targets,
        predictions=metric_functions.logits_to_predictions_multilabel(logits),
        average='macro',
    ),
)



import json
import os
from abc import ABC
from typing import List, Dict, Mapping, Tuple, Optional, TypeVar, Callable, Iterable, Any

import sfu_data_io.helpers as io
import tensorflow as tf
from tensorflow import keras
from tensorflow.keras.callbacks import Callback

import sfu_ml_lib.base.metrics as metric_helpers
from sfu_ml_lib.base.dataset import BatchType
from sfu_ml_lib.base.metrics import Metric, MetricKeyType, NumericType
from sfu_ml_lib.metrics.aggregator import KerasMetric
from sfu_ml_lib.models.schema import Schema


T = TypeVar('T')
W = TypeVar('W')
Z = TypeVar('Z')
BatchFunction = Callable[[], Iterable[BatchType]]


class KerasModel(keras.Model, ABC):
    input_schema: Schema

    @classmethod
    def deconstruct_inputs(cls, features: T, targets: W, weights: Optional[Z] = None) -> Tuple[T, W, Optional[Z]]:
        """
        Converts the outputs of a tf.Dataset into features, targets and optional weights.
        Used in combination with `deconstruct_inputs(*args)`.
        """
        return features, targets, weights

    @classmethod
    def to_metric(cls, label: str) -> MetricKeyType:
        try:
            return Metric(label)
        except ValueError:
            return label

    @classmethod
    def convert_metrics(cls, metrics: Mapping[str, tf.Tensor]) -> Mapping[MetricKeyType, NumericType]:
        return {
            cls.to_metric(label): value.numpy()
            for label, value
            in metrics.items()
        }

    @classmethod
    def from_config_path(cls, path: str, **kwargs) -> 'KerasModel':
        with io.open(os.path.join(path, 'config.json'), mode='r') as json_file:
            config = json.load(json_file)

        config.update(kwargs)

        return cls.from_config(config)

    @classmethod
    def from_config(cls, config: Mapping[str, Any], custom_objects=None) -> 'KerasModel':
        return cls(**config)

    def save_config(self, path: str) -> None:
        with io.open(os.path.join(path, 'config.json'), mode='w') as json_file:
            json.dump(self.get_config(), json_file)

    def get_metrics(self) -> Mapping[str, tf.Tensor]:
        metrics: Dict[str, tf.Tensor] = {}

        for metric in self.metrics_copy:
            if isinstance(metric, KerasMetric):
                metrics.update(metric.get_metrics())
            else:
                metrics[metric.name] = metric.result()

        return metrics

    def reset_metrics(self) -> None:
        for metric in self.metrics_copy:
            metric.reset_states()

    def create_tf_dataset(self, batches: BatchFunction, queue_size: int = -1) -> tf.data.Dataset:
        tensorflow_dataset = tf.data.Dataset.from_generator(
            generator=batches,
            output_types=self.input_schema.get_types(),
            output_shapes=self.input_schema.get_shapes(),
        )

        return tensorflow_dataset.prefetch(queue_size)

    @tf.function
    def _train_dataset(self, dataset: tf.data.Dataset) -> Mapping[str, tf.Tensor]:
        self.reset_metrics()

        for arguments in dataset:
            features, targets, weights = self.deconstruct_inputs(*arguments)

            predictions = self(features, training=True)

            loss = self.loss(targets, predictions, weights)
            if len(self.losses) > 0:
                loss += tf.add_n(self.losses)

            gradients = tf.gradients(loss, self.trainable_weights)
            self.optimizer.apply_gradients(zip(gradients, self.trainable_weights))

            for metric in self.metrics_copy:
                metric(targets, predictions, weights)

        metrics = self.get_metrics()

        return metrics

    def train_dataset(self, batches: BatchFunction) -> Mapping[MetricKeyType, NumericType]:
        """
        Trains a dataset given a generator of input batches. For example:
        ```
        train_dataset(lambda: ((np.random.random((10, 10)), None) for _ in range(100)))
        ```
        If the batching function requires some arguments, they can be curried:
        ```
        train_dataset(functools.partial(dataset.get_batches, batch_size=100))
        ```
        :param batches: A function that receives no arguments and returns an iterator of `BatchType`.
        :return: A dictionary of metrics.
        """
        return self.convert_metrics(self._train_dataset(self.create_tf_dataset(batches)))

    @tf.function
    def _test_dataset(self, dataset: tf.data.Dataset) -> Mapping[str, tf.Tensor]:
        self.reset_metrics()

        for arguments in dataset:
            features, targets, weights = self.deconstruct_inputs(*arguments)

            predictions = self(features, training=False)

            for metric in self.metrics_copy:
                metric(targets, predictions, weights)

        metrics = self.get_metrics()

        return metrics

    def test_dataset(self, batches: BatchFunction) -> Mapping[MetricKeyType, NumericType]:
        """
        Please look at `train_dataset` for details on how to construct `batches`.
        """
        return self.convert_metrics(self._test_dataset(self.create_tf_dataset(batches)))

    def fit_dataset(
            self,
            train_batches: BatchFunction,
            max_num_epochs: int = 1,
            validation_batches: Optional[BatchFunction] = None,
            test_batches: Optional[BatchFunction] = None,
            callbacks: Optional[List[Callback]] = None) -> None:
        """
        Please look at `train_dataset` for details on how to construct `batches`.
        """
        callbacks = callbacks if callbacks else []

        for callback in callbacks:
            callback.set_model(self)
            callback.on_train_begin()

        for epoch in range(max_num_epochs):
            for callback in callbacks:
                callback.on_epoch_begin(epoch)

            train_metrics = self.train_dataset(train_batches)
            metrics = metric_helpers.prefix_metrics_train(train_metrics)

            if validation_batches is not None:
                validation_metrics = self.test_dataset(validation_batches)
                metrics.update(metric_helpers.prefix_metrics_validation(validation_metrics))

            for callback in callbacks:
                callback.on_epoch_end(epoch, metrics)

            if self.stop_training:
                break

        test_metrics = None

        if test_batches is not None:
            test_metrics = metric_helpers.prefix_metrics_test(self.test_dataset(test_batches))

        for callback in callbacks:
            callback.on_train_end(test_metrics)

    def compile(
            self,
            optimizer='rmsprop',
            loss=None,
            metrics=None,
            loss_weights=None,
            sample_weight_mode=None,
            weighted_metrics=None,
            **kwargs) -> None:
        """
        There's a bug in TensorFlow 2.1 that does not set the `metrics` property when compiling.
        So we create our own. This is a hack that should be removed as soon as they fix it.
        """
        super().compile(optimizer, loss, metrics, loss_weights, sample_weight_mode, weighted_metrics, **kwargs)

        self.metrics_copy = metrics if metrics else []



from typing import Optional, Collection, Mapping

import tensorflow as tf
from tensorflow.keras.metrics import Metric

import sfu_ml_lib.base.metrics as metric_helpers
from sfu_ml_lib.base.metrics import NumericType
from sfu_ml_lib.metrics.aggregation.aggregator import Aggregation, MetricAggregator, Statistic


class KerasMetric(Metric):
    def __init__(
            self,
            aggregations: Collection[Aggregation],
            session: tf.compat.v1.Session,
            name: str = 'custom_metric',
            **kwargs,
    ) -> None:

        super().__init__(name, **kwargs)

        self.metric_aggregator = MetricAggregator(aggregations)
        self.session = session

    def update_state(self, y_true: tf.Tensor, y_pred: tf.Tensor, sample_weight: Optional[tf.Tensor] = None) -> None:
        self.metric_aggregator.add_statistics({
            Statistic.TARGET: y_true.eval(session=self.session),
            Statistic.LOGITS: y_pred.eval(session=self.session),
        })

        if sample_weight:
            self.metric_aggregator.add_statistic(Statistic.WEIGHTS, sample_weight.numpy())

    def reset_states(self) -> None:
        self.metric_aggregator.clear()

    def result(self) -> NumericType:
        return next(iter(self.metric_aggregator.get_metrics().values()))

    def get_metrics(self) -> Mapping[str, tf.Tensor]:
        return {
            metric_helpers.to_string(label): tf.convert_to_tensor(value)
            for label, value
            in self.metric_aggregator.get_metrics().items()
        }



from typing import Optional

import tensorflow as tf
from tensorflow.keras.metrics import CategoricalCrossentropy

from sfu_ml_lib.base.metrics import Metric


class MultiClassCrossEntropy(CategoricalCrossentropy):
    def __init__(self, num_classes: int, from_logits: bool):
        super().__init__(name=Metric.CROSS_ENTROPY.value, from_logits=from_logits)
        self.num_classes = num_classes

    def update_state(self, y_true: tf.Tensor, y_pred: tf.Tensor, sample_weight: Optional[tf.Tensor] = None) -> None:
        targets = tf.one_hot(y_true, self.num_classes)
        super().update_state(targets, y_pred, sample_weight)
